import logging

from pyspark.sql import SparkSession


def spark_getOrCreate(
    driver_memory: str = "1g",
    app_name: str = "pyspark Session",
    timezone: str = "UTC",
    aqe: bool = True,
    port: str = "4040",
    config: dict = None,
) -> SparkSession:
    """
    Gets an existing SparkSession or, if there is no existing one, creates a new one
    based on the options set in this function.

    Args:
        driver_memory (str, optional): Amount of memory to use for the driver process,
            i.e. where SparkContext is initialized, in the same format as JVM memory
            strings with a size unit suffix ("k", "m", "g" or "t") (e.g. 512m, 2g).
            Defaults to "1g".
        app_name (str, optional): The name of your application. This will appear in the
            UI and in log data. Defaults to "pyspark Session".
        timezone (str, optional): The ID of session local timezone in the format of
            either region-based zone IDs or zone offsets. This format is also passed to
            the "spark.driver.extraJavaOptions" and "spark.executor.extraJavaOptions".
            Defaults to "UTC".
        aqe (bool, optional): When true, enable adaptive query execution, which
            re-optimizes the query plan in the middle of query execution, based on
            accurate runtime statistics. Defaults to True.
        port (str, optional): Port for your application's dashboard, which shows memory
            and workload data. Defaults to "4040".
        config (dict, optional): Sets a config option. Options set here will overwrite
            any defaults. Defaults to None.

    Returns:
        SparkSession
    """

    sess = (
        SparkSession.builder.appName(app_name)
        .config("spark.driver.memory", driver_memory)
        # -Duser.timezone required in some situations - saving as notes
        .config("spark.driver.extraJavaOptions", f"-Duser.timezone={timezone}")
        .config("spark.executor.extraJavaOptions", f"-Duser.timezone={timezone}")
        .config("spark.sql.session.timeZone", f"{timezone}")
        .config("spark.sql.adaptive.enabled", aqe)
        .config("spark.ui.port", port)
    )

    # User-supplied configs will overwrite default values
    if config is not None:
        for k, v in config.items():
            sess = sess.config(k, v)

    return sess.getOrCreate()


def quiet_py4j():
    """turn down spark logging"""
    logger = logging.getLogger("py4j")
    logger.setLevel(logging.WARN)



from typing import Union, List
from datetime import date

from dateutil.relativedelta import relativedelta
from pyspark.sql import DataFrame, Column
import pyspark.sql.functions as F
from pyspark.sql.window import Window
from pyspark.sql.types import ArrayType, BooleanType, DateType, ShortType, FloatType

from dendri.dataframe import set_column_order


def condense_segments(
    df: DataFrame,
    group_col: Union[str, List[str]],
    start_dt_col: str,
    end_dt_col: str,
    tolerance: int = 1,
) -> DataFrame:
    """
    Combines overlapping start and end dates according to the allowable tolerance for
    each group.

    Args:
        df: Name of dataframe to condense.
        group_col: Columns to group by from df.
        start_dt_col: Name of the start date column from df.
        end_dt_col: Name of the end date column from df.
        tolerance: The allowable number of days between segments.

    Returns:
        Input DataFrame with updated user-supplied start_dt_col, end_dt_col
    """

    # Treat group_col as list in all cases
    if isinstance(group_col, str):
        group_col = [group_col]

    # windowSpec defining group_col + start/end dates
    # - used to determine which segments are overlapping
    win_spec_overlap = (
        Window.partitionBy([F.col(x) for x in group_col])
        .orderBy(F.col(start_dt_col), F.col(end_dt_col))
        .rowsBetween(Window.unboundedPreceding, -1)
    )

    # windowSpec defining group_col + overlapping segments (defined by "_group_id")
    # - used to determine min/max dates across "_group_id"
    win_spec_condense = Window.partitionBy(
        [F.col(x) for x in (group_col + ["_group_id"])]
    )

    return (
        df.withColumn(
            "_group_id",
            F.when(
                F.col(start_dt_col)
                <= F.max(F.date_add(F.col(end_dt_col), tolerance)).over(
                    win_spec_overlap
                ),
                None,
            ).otherwise(F.monotonically_increasing_id()),
        )
        .withColumn(
            "_group_id",
            F.last(F.col("_group_id"), ignorenulls=True).over(
                win_spec_overlap.rowsBetween(Window.unboundedPreceding, 0)
            ),
        )
        .withColumn("_new_start", F.min(F.col(start_dt_col)).over(win_spec_condense))
        .withColumn("_new_end", F.max(F.col(end_dt_col)).over(win_spec_condense))
        .drop(start_dt_col, end_dt_col, "_group_id")
        .withColumnRenamed("_new_start", start_dt_col)
        .withColumnRenamed("_new_end", end_dt_col)
        .orderBy([x for x in (group_col + [start_dt_col] + [end_dt_col])])
    )


def extend_segments(
    df: DataFrame,
    group_col: Union[str, List[str]],
    start_dt_col: str,
    end_dt_col: str,
    tolerance: int = 1,
    retain_shape: bool = True,
) -> DataFrame:
    """
    Extends overlapping start and end dates according to the allowable tolerance for
    each group.

    Args:
        df: Name of dataframe with extendable segments.
        group_col: Columns to group by from df.
        start_dt_col: Name of the start date column from df.
        end_dt_col: Name of the end date column from df.
        tolerance: The allowable number of days between segments.
        retain_shape: Retain original column and row counts.
            Default: True
            False will return distinct `group_col`, `start_dt_col`, `end_dt_col`.

    Returns:
        Input DataFrame with updated user-supplied start_dt_col, end_dt_col
    """

    def extend_array_segments(date_segment_array, tolerance: int):
        output_dates = []
        for idx, dates in enumerate(date_segment_array):
            # First segment: init temp date variables and segment days
            if idx == 0:
                rolling_start_dt = dates[0]
                rolling_end_dt = dates[1]
                rolling_days = (rolling_end_dt - rolling_start_dt).days + 1
            # 2nd+ segment: check if overlaps with rolling segment
            elif dates[0] <= rolling_end_dt + relativedelta(days=tolerance):
                # If overlaps: update rolling segment by adding current segment's days
                current_segment_days = (dates[1] - dates[0]).days + 1
                rolling_days = rolling_days + current_segment_days
                rolling_end_dt = rolling_start_dt + relativedelta(days=rolling_days - 1)
            else:  # This is the start of a segment which does not overlap
                # insert current rolling dates into output_dates
                output_dates.extend([[rolling_start_dt, rolling_end_dt]])

                # Re-initialize dates & days
                rolling_start_dt = dates[0]
                rolling_end_dt = dates[1]
                rolling_days = (rolling_end_dt - rolling_start_dt).days + 1

        # insert last segment into output_dates
        output_dates.extend([[rolling_start_dt, rolling_end_dt]])

        return output_dates

    extend_array_segments_udf = F.udf(
        extend_array_segments, ArrayType(ArrayType(DateType()))
    )

    # Treat group_col as list in all cases
    if isinstance(group_col, str):
        group_col = [group_col]

    extended_segments_sdf = (
        df.groupBy(group_col)
        .agg(
            F.array_sort(F.collect_list(F.array(start_dt_col, end_dt_col))).alias(
                "_input_segments_array"
            ),
        )
        .withColumn(
            "_new_segments",
            extend_array_segments_udf("_input_segments_array", F.lit(tolerance)),
        )
        .select(*group_col, F.explode("_new_segments").alias("_new_segments"))
        .select(
            *group_col,
            F.col("_new_segments")[0].alias(start_dt_col),
            F.col("_new_segments")[1].alias(end_dt_col),
        )
    )

    if retain_shape:
        retained_sdf = (
            df.join(
                extended_segments_sdf.select(
                    *group_col,
                    F.col(start_dt_col).alias("_start_dt_new"),
                    F.col(end_dt_col).alias("_end_dt_new"),
                ),
                group_col,
                "left",
            )
            .filter(
                F.col(start_dt_col).between(
                    F.col("_start_dt_new"), F.col("_end_dt_new")
                )
            )
            .drop(start_dt_col, end_dt_col)
            .withColumnRenamed("_start_dt_new", start_dt_col)
            .withColumnRenamed("_end_dt_new", end_dt_col)
        )
        return set_column_order(retained_sdf, df.columns)
    else:
        return extended_segments_sdf


def covered_days(
    df: DataFrame,
    group_col: Union[str, List[str]],
    segment_start_dt_col: Union[str, Column],
    segment_end_dt_col: Union[str, Column],
    window_start_dt: Union[date, Column],
    window_end_dt: Union[date, Column],
    covered_days_col_name: str = "covered_days",
) -> DataFrame:
    """
    Count days covered by segments within provided window. Days covered is calculated
    without condensing or expanding overlapping segments.

    Args:
        df: Name of dataframe with extendable segments.
        group_col: Columns to group by from df.
        segment_start_dt_col: Segment start date column name.
        segment_end_dt_col: Segment end date column name.
        window_start_dt: Start date of window in which to count covered days.
        window_end_dt: End date of window in which to count covered days.
        covered_days_col_name: Name of column with number of days covered.

    Returns:
        DataFrame:
        {
            ... original columns ...
            `covered_days_col_name` (LongType)
        }
    """

    # group_col always treated like list
    if isinstance(group_col, str):
        group_col = [group_col]

    # convert to F.col() if str was supplied
    if isinstance(segment_start_dt_col, str):
        segment_start_dt_col = F.col(segment_start_dt_col)
    if isinstance(segment_end_dt_col, str):
        segment_end_dt_col = F.col(segment_end_dt_col)

    # convert supplied date to F.lit() if supplied
    if isinstance(window_start_dt, date):
        window_start_dt = F.lit(window_start_dt)
    if isinstance(window_end_dt, date):
        window_end_dt = F.lit(window_end_dt)

    # windowSpec defining group_col + start/end dates
    # - used to determine which segments are overlapping
    win_spec_overlap = (
        Window.partitionBy([F.col(x) for x in group_col])
        .orderBy(segment_start_dt_col, segment_end_dt_col)
        .rowsBetween(Window.unboundedPreceding, -1)
    )

    # windowSpec defining group_col + overlapping segments (defined by "_group_id")
    # - used to determine min/max dates across "_group_id"
    win_spec_condense = Window.partitionBy(
        [F.col(x) for x in (group_col + ["_group_id"])]
    ).orderBy(group_col)

    # windowSpec defining group_col + overlapping segments (defined by "_group_id")
    # - used to determine min/max dates across "_group_id"
    winSpec_first_record = Window.partitionBy(
        [F.col(x) for x in (group_col + ["_group_id_first"] + ["_group_id"])]
    )

    winSpec_group_col = Window.partitionBy(group_col)

    sdf = (
        df.withColumn(
            "_group_id",
            F.when(
                segment_start_dt_col
                <= F.max(F.date_add(segment_end_dt_col, 1)).over(win_spec_overlap),
                None,
            ).otherwise(F.monotonically_increasing_id()),
        )
        .withColumn(
            "_group_id",
            F.last(F.col("_group_id"), ignorenulls=True).over(
                win_spec_overlap.rowsBetween(Window.unboundedPreceding, 0)
            ),
        )
        .withColumn(
            "_group_id_start_dt", F.min(segment_start_dt_col).over(win_spec_condense)
        )
        .withColumn(
            "_group_id_end_dt", F.max(segment_end_dt_col).over(win_spec_condense)
        )
        # Find days covered in specified date range
        .withColumn(
            "_group_id_covered_days",
            F.datediff(
                F.least(F.col("_group_id_end_dt"), window_end_dt),
                F.greatest(F.col("_group_id_start_dt"), window_start_dt),
            )
            + 1,
        )
        # Segments outside window will produce -negative values, zero them
        .withColumn(
            "_group_id_covered_days",
            F.when(F.col("_group_id_covered_days") < 0, 0).otherwise(
                F.col("_group_id_covered_days")
            ),
        )
        # Flag only first record of _group_id
        .withColumn(
            "_group_id_first",
            F.when(F.row_number().over(win_spec_condense) == 1, True).otherwise(False),
        )
        # Sum across only the first segment of each _group_id
        .withColumn(
            covered_days_col_name,
            F.sum(F.col("_group_id_covered_days")).over(winSpec_first_record),
        )
        .withColumn(
            covered_days_col_name,
            F.sum(
                F.when(
                    F.col("_group_id_first"), F.col(covered_days_col_name)
                ).otherwise(0)
            ).over(winSpec_group_col),
        )
    ).drop(
        "_group_id",
        "_group_id_start_dt",
        "_group_id_end_dt",
        "_group_id_covered_days",
        "_group_id_first",
    )

    return sdf


def first_event_in_x_days(
    df: DataFrame,
    group_col: Union[str, List[str]],
    start_dt_col: str,
    days: int,
) -> DataFrame:
    """
    Identify the first valid event in an x-day period. Identify invalid events that are
    within x-days of the last valid event.

    Args:
        df: Input DataFrame
        group_col: Columns to group by from input DataFrame
        start_dt_col: Start date column from input DataFrame
        days: Number of days to check between events

    Returns:
        Input DataFrame with BooleanType column 'valid_event' for showing if the event
        was the first in an x-day period.
            Columns: {
                group_col
                start_dt_col (DateType())
                valid_event  (BooleanType())
            }
    """

    def first_event_func(dates, tolerance):
        comp_dt = dates[0]
        keep_dt = []
        for i, dt in enumerate(dates):
            if i == 0:
                keep_dt.append(True)
            elif dt <= comp_dt + relativedelta(days=tolerance):
                keep_dt.append(False)
            else:
                keep_dt.append(True)
                comp_dt = dt

        return keep_dt

    first_event_udf = F.udf(first_event_func, ArrayType(BooleanType()))

    if isinstance(group_col, str):
        group_col = [group_col]

    return (
        df.groupBy(group_col)
        # Convert long column to sorted array per group
        .agg(F.array_sort(F.collect_list(start_dt_col)).alias("_temp_dt"))
        # Iterate over array per group instead of iterating over RDD rows
        .withColumn("_temp_status", first_event_udf("_temp_dt", F.lit(days)))
        .select(
            *group_col,
            F.explode(F.arrays_zip("_temp_dt", "_temp_status")).alias("_temp_map"),
        )
        .select(
            *group_col,
            F.col("_temp_map")["_temp_dt"].alias(start_dt_col),
            F.col("_temp_map")["_temp_status"].alias("valid_event"),
        )
    )


def count_events_by_period(
    df: DataFrame,
    group_col: Union[str, List[str]],
    count_col: str,
    dt_col: str,
    max_dt: Union[str, date],
    period: str,
    num_periods: int,
    count_unique: bool = False,
) -> DataFrame:
    """
    Count the number of events per period per group.

    Args:
        df: Name of input DataFrame
        group_col: Name of column(s) to group by when counting
        count_col: Name of column to count
        dt_col: Name of date column for determining periods
        max_dt: End date of final period
        period: Period type to use. Accepted values are 'years', 'months', 'days', and
            'weeks'
        num_periods: Number of period to count (going backwards from max_dt)
        count_unique: Count unique values. Defaults to False.

    Returns:
        Number of events per period per group.
            Columns: {
                group_col
                event_cnt_prd0
                ...
                event_cnt_prd{num_periods - 1}
            }
    """

    if isinstance(max_dt, str):
        max_dt = date.fromisoformat(max_dt)

    end_dt = max_dt

    period_col_list = []
    period = period.lower()

    # Create columns for each period populated with the count col
    for i in range(1, num_periods + 1):
        if period in ("y", "year", "years"):
            offset = relativedelta(years=i)
        if period in ("m", "months", "month"):
            offset = relativedelta(months=i)
        elif period in ("d", "day", "days"):
            offset = relativedelta(days=i)
        elif period in ("w", "weeks", "week"):
            offset = relativedelta(weeks=i)
        else:
            raise ValueError(
                "period only supports the values of 'years', 'months', 'days', and "
                "'weeks'"
            )

        start_dt = max_dt - offset + relativedelta(days=1)
        col_name = f"event_cnt_prd{i - 1}"
        period_col_list.append(col_name)

        df = df.withColumn(
            col_name,
            F.when(
                F.col(dt_col).between(F.lit(start_dt), F.lit(end_dt)), F.col(count_col)
            ),
        )

        end_dt = max_dt - offset

    if count_unique:
        agg_func = [
            F.countDistinct(col).cast(ShortType()).alias(col) for col in period_col_list
        ]
    else:
        agg_func = [
            F.count(col).cast(ShortType()).alias(col) for col in period_col_list
        ]

    return df.groupBy(group_col).agg(*agg_func)


def age(
    dob_col: str,
    age_dt: Union[str, date, Column],
    floor_sw: bool = True,
) -> Column:
    """
    Calculate age.

    Args:
        dob_col: Column name containing date of birth.
        age_dt: Date to calculate age.
        floor_sw: Return age as whole number, rounded down. Defaults to True.

    Returns:
        Age as of the specified date. Returns FloatType if round is False. Otherwise
            returns ShortType.
    """
    if isinstance(age_dt, str):
        age_dt = date.fromisoformat(age_dt)

    if isinstance(age_dt, date):
        age_dt = F.lit(age_dt)

    age = (F.months_between(age_dt, dob_col) / 12).cast(FloatType())

    if floor_sw:
        age = F.floor(age).cast(ShortType())

    return age



import re
import json
import types
import functools
from pathlib import Path
from typing import Union, List
from collections import namedtuple

import pyspark.sql.functions as F
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.column import Column
from pyspark.sql.types import (
    ArrayType,
    StringType,
    StructType,
    MapType,
    BinaryType,
    BooleanType,
    DateType,
    NullType,
    NumericType,
    TimestampType,
    UserDefinedType,
)


def validate_config_path(path: Union[str, Path]):
    """
    Pending deprecation.
    """
    path = Path(path)

    # validate user-supplied path
    #   - Must end with '.config.json'
    name_parts = path.resolve().name.split(".")
    if name_parts[-1] != "json" or name_parts[-2] != "config":
        raise ValueError(
            "path to dataframe config file must end with '.config.json'\n"
            f"path: {path.resolve()}"
        )


def show_by_type(sdf: DataFrame, dtype: str, *args, **kwargs):
    """
    .show() columns of provided dtype

    Engine: Spark
    """
    columnList = [item[0] for item in sdf.dtypes if item[1].startswith(dtype)]
    sdf.select(columnList).show(*args, **kwargs)


def namedtuple_from_schema(schema: StructType, name: str):
    """
    Convert a PySpark schema into a named tuple

    Args:
        schema: PySpark schema
        name: Name of namedtuple

    Returns:
        namedtuple representation of PySpark schema
    """
    return namedtuple(name, schema.names, defaults=(None,) * len(schema.names))


def cols_to_array(*cols, remove_na: bool = True) -> Column:
    """
    Create a column of ArrayType() from user-supplied column list.

    Args:
        cols: columns to convert into array.
        remove_na (optional): Remove nulls from array. Defaults to True.

    Returns:
        Column of ArrayType()
    """
    if remove_na:
        return F.filter(F.array(*cols), lambda x: x.isNotNull())
    else:
        return F.array(*cols)


def cols_to_dict(df: DataFrame, key_col_name: str, value_col_name: str) -> dict:
    """
    Creates a dictionary from pair of columns.

    Args:
        df: PySpark DataFrame
        key_col_name: Column name to use as dict key
        value_col_name: Column name to use as dict value

    Returns:
        A dict of the two DataFrame columns
    """
    k, v = key_col_name, value_col_name
    return {x[k]: x[v] for x in df.select(k, v).collect()}


def dict_to_map(mapping: dict) -> Column:
    """
    Create a PySpark map literal from dict of k-v pairs.

    Args:
        mapping: Mapping dict

    Returns:
        PySpark Column of MapType()
    """
    literal_list = []

    for k, v in mapping.items():
        literal_list.append(F.lit(k))
        if isinstance(v, (str, int, float, bool)):
            literal_list.append(F.lit(v))
        elif isinstance(v, (list, tuple, set)):
            literal_list.append(F.array(*[F.lit(x) for x in v]))

    return F.create_map(*literal_list)


def map_column(col_name: str, mapping: Union[Column, dict]) -> Column:
    """
    Apply mapping expression to column.

    Args:
        col_name: Name of column to map values
        mapping: MapType() column or dict with desired mapping

    Returns:
        Column with mapped values
    """
    if isinstance(mapping, dict):
        mapping = dict_to_map(mapping)

    return mapping[F.col(col_name)]


class DataFrameMissingColumnError(ValueError):
    """Raise this when there's a missing DataFrame column"""


class DataFrameMissingStructFieldError(ValueError):
    """Raise this when there's a missing DataFrame StructField"""


class DataFrameProhibitedColumnError(ValueError):
    """Raise this when a DataFrame includes prohibited columns"""


def validate_presence_of_columns(df: DataFrame, required_col_names: List[str]):
    """
    Validate DataFrame contains listed columns.

    Args:
        df: PySpark DataFrame to check columns
        required_col_names: List of column names to validate

    Raises:
        DataFrameMissingColumnError: If DataFrame is missing a required column
    """
    missing_cols = [x for x in required_col_names if x not in df.columns]

    error_message = (
        "The following columns are not included in the DataFrame:"
        f" {', '.join(missing_cols)}"
    )

    if missing_cols:
        raise DataFrameMissingColumnError(error_message)


def validate_schema(df: DataFrame, required_schema: StructType):
    """
    Raises Error if df schema does not match required schema.

    Args:
        df: PySpark DataFrame to validate schema
        required_schema: Required schema

    Example:
        data = [("jose", 1), ("li", 2), ("luisa", 3)]
        source_df = spark.createDataFrame(data, ["name", "age"])
        required_schema = StructType(
            [
                StructField("name", StringType(), True),
                StructField("city", StringType(), True),
            ]
        )
        # Raises Error, as "city" does not exist, but "age" does.
        validate_schema(source_df, required_schema)

    Raises:
        DataFrameMissingStructFieldError: If DataFrame is missing a StructField
    """
    missing_struct_fields = [x for x in required_schema if x not in df.schema]

    error_message = (
        "The following StructFields are not included in the DataFrame:"
        f" {missing_struct_fields}"
    )

    if missing_struct_fields:
        raise DataFrameMissingStructFieldError(error_message)


def validate_absence_of_columns(df: DataFrame, prohibited_col_names: List[str]):
    """
    Validate DataFrame does not contain listed columns.

    Args:
        df: PySpark DataFrame to check columns
        prohibited_col_names: List of column names to validate

    Raises:
        DataFrameProhibitedColumnError: If DataFrame is includes a prohibited column
    """
    extra_cols = [x for x in prohibited_col_names if x in df.columns]

    error_message = (
        "The following columns are not allowed to be included in the DataFrame:"
        f" {', '.join(extra_cols)}"
    )

    if extra_cols:
        raise DataFrameProhibitedColumnError(error_message)


def string_to_column_name(x: str) -> str:
    """
    Converts non-alphanumeric characters to "_" and lowercases string. Repeated
    non-alphanumeric characters are replaced with a single "_".

    Args:
        x: String to convert to column name

    Returns:
        Re-formatted string for using as a PySpark DataFrame column name
    """
    return re.sub("[^0-9a-zA-Z]+", "_", x).lower()


def stage_dataframe_to_disk(
    cached_file_path: Union[str, Path], overwrite: bool = False
):
    """
    Stage returned DataFrame to disk. Decorated function must return a Spark DataFrame.

    Args:
        cached_file_path: path to desired parquet location on disk.
        overwrite: overwrite existing parquet if exists.

    Example
    -------
    ```
    @stage_dataframe_to_disk("path-to-file/file.parquet")
    def very_important_sdf():
        data = [(1,), (2,)]
        sdf = spark.createDataFrame(data, ["col"])
        return sdf

    # DataFrame is written to disk @ path-to-file/file.parquet
    very_important_sdf()
    ```
    """

    def decorator_to_disk(func):
        @functools.wraps(func)
        def wrapper_to_disk(
            *args, cached_file_path=cached_file_path, overwrite=overwrite, **kwargs
        ):
            cached_file_path = Path(cached_file_path)
            spark = SparkSession.getActiveSession()
            cached_file_path = Path(cached_file_path)
            if cached_file_path.exists() is False or overwrite is True:
                func(*args, **kwargs).write.parquet(
                    str(cached_file_path), mode="overwrite"
                )
            return spark.read.parquet(str(cached_file_path))

        return wrapper_to_disk

    return decorator_to_disk


def data_template_from_schema(schema: StructType):
    """Returns a hand-fillable template based on supplied schema."""
    header_string = ""
    output = "("
    for field in schema.fields:
        if isinstance(field.dataType, StringType):
            output += '"String", '
        elif isinstance(field.dataType, DateType):
            output += 'date.fromisoformat(""), '
        elif isinstance(field.dataType, TimestampType):
            output += 'datetime.fromisoformat(""), '
        elif isinstance(field.dataType, ArrayType):
            output += '["String?"], '
        elif isinstance(field.dataType, (MapType, StructType)):
            output += '{"": ""}, '
        elif isinstance(field.dataType, BinaryType):
            output += 'b"", '
        elif isinstance(field.dataType, BooleanType):
            output += "True False, "
        elif isinstance(field.dataType, NumericType):
            output += f"{field.dataType}, "
        elif isinstance(field.dataType, NullType):
            output += "None, "
        elif isinstance(field.dataType, UserDefinedType):
            output += "UDT(), "
        else:
            # Something numeric
            output += "Error, "
        header_string += field.name + ", "
    header_string = header_string[:-2]
    output = output[:-2] + ")"
    return header_string + "\n" + output


def set_column_order(
    df: DataFrame, column_order: list, remove_unlisted: bool = False
) -> DataFrame:
    """
    Set the column order for a DataFrame to a desired order. DataFrame colums not in
    column_order will be placed at the end of the DataFrame.

    Args:
        df: DataFrame to reorder columns.
        column_order: List of column names in desired order.
        remove_unlisted: Remove columns from DataFrame if not in supplied column list.
            Default: False.

    Returns:
        DataFrame with columns in desired order.
    """

    # Type check the inputs
    if not isinstance(df, DataFrame):
        raise TypeError("df must be a Spark DataFrame.")
    if not isinstance(column_order, list):
        raise TypeError("column_order must be a list" f"{column_order} was supplied")
    if not isinstance(remove_unlisted, bool):
        raise TypeError("remove_unlisted must be a bool")

    # List of columns in both df & column_order
    ordered_columns_intersect = [col for col in column_order if col in set(df.columns)]

    # columns in df that were not provided to function
    in_df_not_provided = [col for col in df.columns if col not in set(column_order)]

    if remove_unlisted is False:
        ordered_columns = ordered_columns_intersect + in_df_not_provided
    elif remove_unlisted is True:
        ordered_columns = ordered_columns_intersect

    return df.select(ordered_columns)


def rename_by_dict(df: DataFrame, renames: dict) -> DataFrame:
    """
    Renames DataFrame columns according to dictionary

    Args:
        df: PySpark DataFrame
        renames: Dictionary of renames in {existing: new} format

    Returns:
        DataFrame with column names updated
    """

    col_list = []

    for column in df.columns:
        if column in renames.keys():
            col_list.append(F.col(column).alias(renames[column]))
        else:
            col_list.append(F.col(column))

    return df.select(col_list)


def outer_union_corr(*dfs: DataFrame) -> DataFrame:
    """
    Returns a new DataFrame containing the union of all rows and columns from passed
    DataFrames. This is equivalent to OUTER UNION CORR in SQL. To do a SQL-style set
    union (that does deduplication of elements), use this function followed by
    distinct(). The difference between this function and unionByName() is that this
    function returns all columns from all DataFrames.

    This is equivalent to chaining unionByName() with allowMissingColumns=True.

    Args:
        dfs: Sequence of DataFrames to union

    Returns:
        DataFrame
    """
    for i, df in enumerate(dfs):
        # For first df, reassign to _df. Otherwise overwrite and union with _df
        if i == 0:
            _df = df
        else:
            _df = _df.unionByName(df, allowMissingColumns=True)

    return _df


def col_to_set(df: DataFrame, col: str) -> set:
    """
    Returns a set from specified column of DataFrame

    Args:
        df: PySpark DataFrame
        col: Name of column

    Returns:
        A set
    """
    return set(df.select(col).distinct().toPandas()[col])


def col_to_list(df: DataFrame, col: str) -> list:
    """
    Returns a list from specified column of DataFrame

    Args:
        df: PySpark DataFrame
        col: Name of column

    Returns:
        A list
    """
    return list(df.select(col).toPandas()[col])


def array_isin(
    array_col: str, values: Union[str, list], require_all: bool = False
) -> Column:
    """
    Check if one or more array values are in user-supplied list.

    Args:
        array_col: Array column name
        values: List of values to check
        require_all (optional): If True, requires the array to contain all values.
            Defaults to False.

    Returns:
        Column of BooleanType()
    """
    if isinstance(values, str):
        condition = F.array_contains(F.col(array_col), F.lit(values))
    else:
        condition = F.array_contains(F.col(array_col), F.lit(values[0]))

    if len(values) > 1:
        for v in values[1:]:
            if require_all:
                condition = condition & (F.array_contains(F.col(array_col), F.lit(v)))
            else:
                condition = condition | (F.array_contains(F.col(array_col), F.lit(v)))

    return condition


def saveParquetTable(
    self,
    table_name: str,
    file_path: Union[str, Path],
    partition_cols: Union[List[str], str] = None,
    bucket_cols: Union[List[str], str] = None,
    bucket_size: int = None,
    sort_cols: Union[List[str], str] = None,
    write_mode="overwrite",
) -> None:
    """
    Saves a DataFrame using the saveAsTable method and writes necessary metadata to the
    parquet file as _dendri_metadata. This metadata is used when calling
    `readParquetTable` to return a DataFrame that retains information about bucketing.

    Args:
        table_name: Name of table to create
        file_path: Path to save table
        partition_cols (optional): Columns to partition by. Defaults to None.
        bucket_cols (optional): Columns to bucket by. Defaults to None.
        bucket_size (optional): Number of buckets per partition. Defaults to None.
        sort_cols (optional): Columns to sort each bucket by. Only supports ascending
            order. Defaults to None.
        write_mode (optional): Mode to use when calling `saveAsTable`. Defaults to
            "overwrite".
    """
    if isinstance(file_path, Path):
        file_path = str(file_path)

    if partition_cols is not None:
        save_builder = self.repartition(partition_cols).write.partitionBy(
            partition_cols
        )
    else:
        save_builder = self.write

    if bucket_cols is not None:
        save_builder = save_builder.bucketBy(bucket_size, bucket_cols)

    if sort_cols is not None:
        save_builder = save_builder.sortBy(sort_cols)

    save_builder.option("path", file_path).saveAsTable(
        table_name, format="parquet", mode=write_mode
    )

    # Create dictionary of metadata
    meta_dict = {
        "partition_cols": partition_cols,
        "bucket_cols": bucket_cols,
        "bucket_size": bucket_size,
        "sort_cols": sort_cols,
        "schema": self.dtypes,
    }

    # Write metadata to parquet file
    with open(f"{file_path}/_dendri_metadata", "w") as f:
        json.dump(meta_dict, f, indent=4)


def _convert_dtypes_to_str(x: list) -> str:
    return ", ".join([f"{i[0]} {i[1]}" for i in x])


def _convert_list_to_str(x: Union[str, list]) -> str:
    if isinstance(x, list):
        x = ", ".join(x)
    else:
        x = str(x)

    return x


def _sql_builder(
    file: str,
    table: str,
    schema: str,
    partition_cols: str,
    bucket_cols: str,
    bucket_size: str,
    sort_cols: str,
) -> str:
    create_table = f"CREATE TABLE IF NOT EXISTS {table} ({schema}) USING PARQUET "

    if partition_cols != "None":
        partition_by = f"PARTITIONED BY ({partition_cols}) "
    else:
        partition_by = ""

    if bucket_cols != "None":
        clustered_by = f"CLUSTERED BY ({bucket_cols}) "
    else:
        clustered_by = ""

    if sort_cols != "None":
        sorted_by = f"SORTED BY ({sort_cols}) "
    else:
        sorted_by = ""

    if bucket_size != "None":
        into_buckets = f"INTO {bucket_size} BUCKETS "
    else:
        into_buckets = ""

    location = f"LOCATION '{file}'"

    sql = (
        create_table + partition_by + clustered_by + sorted_by + into_buckets + location
    )

    return sql


def readParquetTable(
    self, file_path: Union[str, Path], table_name: str = None
) -> DataFrame:
    """
    Reads parquet files using the PySpark Table API to return a DataFrame with bucketing
    information retained. Bucketing info is only kept when writing the DataFrame using
    the `saveParquetTable` method.

    Args:
        file_path: Path to parquet file
        table_name: Name of table to create when reading parquet file

    Returns:
        DataFrame
    """
    if table_name is None:
        table_name = Path(file_path).stem

    if isinstance(file_path, Path):
        file_path = str(file_path)

    meta_file = f"{file_path}/_dendri_metadata"

    if Path(meta_file).exists():
        with open(meta_file, "r") as f:
            meta_dict = json.load(f)

        # Extract metadata and convert to strings needed for CREATE TABLE SQL
        partition_cols = _convert_list_to_str(meta_dict["partition_cols"])
        bucket_cols = _convert_list_to_str(meta_dict["bucket_cols"])
        bucket_size = _convert_list_to_str(meta_dict["bucket_size"])
        sort_cols = _convert_list_to_str(meta_dict["sort_cols"])
        schema = _convert_dtypes_to_str(meta_dict["schema"])

        # Build the CREATE TABLE SQL
        sql = _sql_builder(
            file_path,
            table_name,
            schema,
            partition_cols,
            bucket_cols,
            bucket_size,
            sort_cols,
        )

        self.sql(sql)

        # Need to register partitions with Hive manually
        if partition_cols != "None":
            self.sql(f"MSCK REPAIR TABLE {table_name}")

        df = self.table(table_name)
    else:
        df = self.read.parquet(file_path)

    return df


DataFrame.saveParquetTable = saveParquetTable
SparkSession.readParquetTable = readParquetTable


def get_dataframe(obj) -> DataFrame:
    """Attempts to return a Spark DataFrame from supplied reference."""
    spark = SparkSession.getActiveSession()

    # Cast Path to str
    if isinstance(obj, Path):
        obj = str(obj)

    if isinstance(obj, str):
        return spark.readParquetTable(obj)
    if isinstance(obj, DataFrame):
        return obj
    if isinstance(obj, types.FunctionType):
        return obj()
    if isinstance(obj, types.MethodType):
        # NOTE(Wes): Not 100% sure on this one.
        return obj()
    # Return whatever it is and let the end user figure it out (None, perhaps?)
    return obj


def melt(
    df: DataFrame,
    melt_cols: Union[List[str], str],
    id_cols: Union[List[str], str, None] = None,
    col_name: str = "col",
    val_name: str = "val",
) -> DataFrame:
    """
    Melt a DataFrame from wide to long format

    Args:
        df (DataFrame): DataFrame to melt
        melt_cols (Union[List[str], str]): Columns to melt
        id_cols (Union[List[str], str, None], optional): ID columns to not melt.
            Defaults to None.
        col_name (str, optional): Name of column containing original column names.
            Defaults to "col".
        val_name (str, optional): Name of column containing original values. Defaults to
            "val".

    Returns:
        DataFrame: Melted DataFrame
    """
    if isinstance(id_cols, str):
        id_cols = [id_cols]
    if isinstance(melt_cols, str):
        melt_cols = [melt_cols]

    exploded_cols: Column = F.explode(
        F.array(
            *[
                F.struct(F.lit(c).alias(col_name), F.col(c).alias(val_name))
                for c in melt_cols
            ]
        )
    ).alias("__col")

    _df: DataFrame

    if id_cols is None:
        _df = df.select(exploded_cols).select(F.col("__col.*"))
    else:
        _df = df.select(*id_cols, exploded_cols).select(*id_cols, F.col("__col.*"))

    return _df


def merge_map_arrays(*cols: Union[str, Column]) -> Column:
    """
    Merge N-maps where the values are arrays by unioning them together.

    Args:
        cols (Union[str, Column]): List of columns of type Map(Any, Array) to combine

    Returns:
        Column: Merged maps of type Map(Any, Array)
    """

    if len(cols) < 2:
        raise ValueError(
            "Not enough values passed. merge_map_arrays must require at least 2 cols"
        )

    for i, col in enumerate(cols):
        # Ensure columns do not contain nulls--must be an empty map
        col = F.coalesce(col, F.create_map())

        if i == 0:
            merge_col = col
        else:
            # Merge maps based on key
            # Values are arrays and will be unioned together
            # map_zip_with defaults to null if missing key--coalesce with an empty array
            merge_col = F.map_zip_with(
                merge_col,
                col,
                lambda k, v1, v2: F.array_union(
                    F.coalesce(v1, F.array()), F.coalesce(v2, F.array())
                ),
            )

    return merge_col



# Make modules available
from . import dataframe
from . import dates
from . import utils

# Shortcut specific functionality
from dendri.dataframe import SparkSession, DataFrame

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



import tempfile
from shutil import rmtree
from contextlib import contextmanager

import pytest

from dendri.utils import spark_getOrCreate, quiet_py4j


@pytest.fixture(scope="session")
def spark_context(request):
    """
    fixture for creating a spark sql session

    Args:
        request: pytest.FixtureRequest object
    """
    spark_context = spark_getOrCreate(
        app_name="pytest-spark-fixture",
        config={
            "spark.sql.shuffle.partitions": 1,
            "spark.default.parallelism": 1,
            "spark.rdd.compress": False,
            "spark.shuffle.compress": False,
            # Note(Wes): I didn't see any decrease in runtime after making the below
            #   changes, though they're listed as helpful on the pytest-spark repo:
            #   https://github.com/malexer/pytest-spark
            #   Keeping the options commented out and available as future options
            # "spark.dynamicAllocation.enabled": "false",
            # "spark.executor.cores": 1,
            # "spark.executor.instances": 1,
            # "spark.io.compression.codec": "lz4",
            # "spark.sql.catalogImplementation": "hive",
        },
    )
    quiet_py4j()
    try:
        yield spark_context
    finally:
        spark_context.stop()


@contextmanager
def ensure_clean_dir():
    """
    Get a temporary directory path and agrees to remove on close.
    Yields
    ------
    Temporary directory path
    """
    directory_name = tempfile.mkdtemp(suffix="")
    try:
        yield directory_name
    finally:
        try:
            rmtree(directory_name)
        except OSError:
            pass
